/*
 * Copyright (c) 2003
 * Gerasimos Dimitriadis, Aristotle University of Thessaloniki.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GERASIMOS DIMITRIADIS AND THE ARISTOTLE
 * UNIVERSITY OF THESSALONIKI ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * SHALL GERASIMOS DIMITRIADIS OR THE ARISTOTLE UNIVERSITY OF THESSALONIKI
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <sys/socket.h>
#include <sys/ctype.h>

#include <net/if.h>
#include <net/ethernet.h>

#include <netgraph/ng_message.h>
#include <netgraph/netgraph.h>
#include <netgraph/ng_parse.h>
#include <netgraph/ng_ether.h>
#include <netgraph/ng_eiface.h>

#include "ng_daphne.h"
#include "ng_daphne_macros.h"

#ifdef NG_SEPARATE_MALLOC
MALLOC_DEFINE(M_NETGRAPH_DAPHNE, "netgraph_daphne", "netgraph daphne node");
#else
#define M_NETGRAPH_DAPHNE M_NETGRAPH
#endif

/*
 * Private data for daphne node
 */
struct ng_daphne_private
{
	hook_p				 uplink, downlink, info, pass;
	node_p				 node;
	struct callout		 handle;

	struct rtElem		*spareRtElems, *heardNeighbours;
	struct rowHead		*spareRowHeads, *routingTable, *myRowHead;
	u_int8_t			 numSpareRowHeads, numSpareRtElems;
	u_int8_t			 numRowHeads, numHeardNeighbours;
	u_int8_t			 updateCounter;
};

typedef struct ng_daphne_private *priv_p;

/* Netgraph node methods */
static ng_constructor_t		ng_daphne_constructor;
static ng_rcvmsg_t			ng_daphne_rcvmsg;
static ng_shutdown_t		ng_daphne_shutdown;
static ng_newhook_t			ng_daphne_newhook;
static ng_rcvdata_t			ng_daphne_rcvdata;
static ng_connect_t			ng_daphne_connect;
static ng_disconnect_t		ng_daphne_disconnect;

/* Other internal functions */
static void		ng_daphne_timeout (void *arg);
static int		ng_daphne_transmit_beacon (priv_p priv);
static int		ng_daphne_process_beacon (priv_p priv,
					struct ng_item *item, struct mbuf *m);
static int		ng_daphne_transmit_update (priv_p priv);
static int		ng_daphne_process_update (priv_p priv,
					struct ng_item *item, struct mbuf *m);
static int		ng_daphne_add_heard_neighbour (priv_p priv, u_int8_t *addr);
static int		ng_daphne_add_neighbour (priv_p priv, u_int8_t *addr);
static int		ng_daphne_calc_shortest_paths (priv_p priv);
static int		ng_daphne_process_data_from_downlink (priv_p priv,
					struct ng_item *item, struct mbuf *m);
static int		ng_daphne_process_data_from_uplink (priv_p priv,
					struct ng_item *item, struct mbuf *m);
static int		ng_daphne_transmit_rt_info (priv_p priv);


/*
 * Daphne Command List
 */
static const struct ng_cmdlist ng_daphne_cmdlist[] = {
	{
	  NGM_DAPHNE_COOKIE,
	  NGM_DAPHNE_SET_ENADDR,
	  "setenaddr",
	  &ng_ether_enaddr_type,
	  NULL
	},
	{
	  NGM_DAPHNE_COOKIE,
	  NGM_DAPHNE_GET_ENADDR,
	  "getenaddr",
	  NULL,
	  &ng_ether_enaddr_type
	},
	{ 0 }
};


/*
 * Node type descriptor
 */
static struct ng_type ng_daphne_typestruct = {
	NG_ABI_VERSION,
	NG_DAPHNE_NODE_TYPE,
	NULL,
	ng_daphne_constructor,
	ng_daphne_rcvmsg,
	ng_daphne_shutdown,
	ng_daphne_newhook,
	NULL,
	ng_daphne_connect,
	ng_daphne_rcvdata,
	ng_daphne_disconnect,
	ng_daphne_cmdlist,
};

NETGRAPH_INIT(daphne, &ng_daphne_typestruct);

/* Depend on ng_ether, the downlink hook connects only to
   ng_ether nodes. Also, we use the Ethernet parse type */
MODULE_DEPEND(ng_daphne, ng_ether, 1, 1, 1);


/*
 * Node constructor
 */
static int
ng_daphne_constructor (node_p node)
{
	priv_p			 priv;
	int				 i;
	struct rowHead	*p_tempRowHead;
	struct rtElem	*p_tempRtElem;

	MALLOC(priv, priv_p, sizeof(*priv), M_NETGRAPH_DAPHNE, M_NOWAIT | M_ZERO);
	if (priv == NULL)
	{
		return (ENOMEM);
	}

	/* Construct some spare rowHeads */
	for (i = 0; i < MIN_SPARE_ROWHEADS; i++)
	{
		MALLOC (p_tempRowHead, struct rowHead*, sizeof (struct rowHead), M_NETGRAPH_DAPHNE, M_NOWAIT);
		if (p_tempRowHead == NULL)
		{
			/* No memory - abort */
			FREE_CHAIN (priv->spareRowHeads, rowHead);
			FREE (priv, M_NETGRAPH_DAPHNE);
			return (ENOMEM);
		}
		else
		{
			/* Add new rowHead to list */
			bzero (p_tempRowHead, sizeof (struct rowHead));
			ADD_TO_CHAIN (priv->spareRowHeads, p_tempRowHead, priv->numSpareRowHeads);
		}
	}

	/* Construct some spare rtElems */
	for (i = 0; i < MIN_SPARE_RTELEMS; i++)
	{
		MALLOC (p_tempRtElem, struct rtElem*, sizeof (struct rtElem), M_NETGRAPH_DAPHNE, M_NOWAIT);
		if (p_tempRtElem == NULL)
		{
			/* No memory - abort */
			FREE_CHAIN (priv->spareRtElems, rtElem);
			FREE_CHAIN (priv->spareRowHeads, rowHead);
			FREE (priv, M_NETGRAPH_DAPHNE);
			return (ENOMEM);
		}
		else
		{
			/* Add new rtElem to list */
			bzero (p_tempRtElem, sizeof (struct rtElem));
			ADD_TO_CHAIN (priv->spareRtElems, p_tempRtElem, priv->numSpareRtElems);
		}
	}

	NG_NODE_FORCE_WRITER(node);
	NG_NODE_SET_PRIVATE(node, priv);

	priv->node = node;

	priv->updateCounter = 1;

	/* Set function that handles timer timeouts, but do not start the timers yet.
	   Wait for the downlink hook to connect to start the timer */
	callout_init(&priv->handle, 0);

	return (0);
}


/*
 * A hook ask to be connected to us. Initially say ok. We will check later if
 * a ether node asks to be connected to our downlink hook
 */
static int
ng_daphne_newhook(node_p node, hook_p hook, const char *name)
{
	const priv_p priv = NG_NODE_PRIVATE(node);

	if (!strcmp(name, NG_DAPHNE_HOOK_DOWNSTREAM))
	{
		if (priv->downlink)
			return (EISCONN);

		priv->downlink = hook;
		return (0);
	}

	if (!strcmp(name, NG_DAPHNE_HOOK_PASS))
	{
		if (priv->pass)
			return (EISCONN);

		priv->pass = hook;
		return (0);
	}

	if (!strcmp(name, NG_DAPHNE_HOOK_UPSTREAM))
	{
		if (priv->uplink)
			return (EISCONN);

		priv->uplink = hook;
		return (0);
	}

	if (!strcmp(name, NG_DAPHNE_HOOK_INFO))
	{
		if (priv->info)
			return (EISCONN);

		priv->info = hook;
		return (0);
	}

	return (EINVAL);
}


/*
 * Validate hook connection. Accepted downlink hooks only if the peer
 * node is of ether type.
 */
static int
ng_daphne_connect (hook_p hook)
{
	struct ng_mesg *msg;
	priv_p	priv = NG_NODE_PRIVATE(NG_HOOK_NODE(hook));
	node_p	node = NG_HOOK_NODE(hook);
	int error;

	if (hook == priv->downlink)
	{
		/* Check if the node asking to be connected to downlink is of type ether */
		if (strcmp(NG_PEER_NODE(hook)->nd_type->name, "ether") != 0)
		{
			/* If not ether, refuse to be connected to this hook */
			return (EINVAL);
		}

		/* The peer node is ether, construct a message to ask about its hardware address */
		NG_MKMESSAGE (msg, NGM_ETHER_COOKIE, NGM_ETHER_GET_ENADDR, 0, M_NOWAIT);

		if (msg == NULL)
			return (ENOMEM);

		/* Send message to this node */
		NG_SEND_MSG_ID (error, node, msg, NG_PEER_NODE(hook)->nd_ID, 0);

		/* Accept connection */
		return (0);
	}
	else if (hook == priv->uplink)
		{
			if (strcmp (NG_PEER_NODE(hook)->nd_type->name, "eiface") != 0)
			{
				/* If not eiface, refuse to be connected to this hook */
				return (EINVAL);
			}

			/* The peer node is eiface. If we know out hardware address, let it know */
			if (priv->myRowHead)
			{
				struct ng_eiface_par *eiAddr;

				NG_MKMESSAGE (msg, NGM_EIFACE_COOKIE, NGM_EIFACE_SET, sizeof (struct ng_eiface_par), M_NOWAIT);

				if (msg == NULL)
					return (ENOMEM);

				/* Fill up msg data */
				eiAddr = (struct ng_eiface_par *)(msg->data);

				eiAddr->oct0 = priv->myRowHead->etherAddr[0];
				eiAddr->oct1 = priv->myRowHead->etherAddr[1];
				eiAddr->oct2 = priv->myRowHead->etherAddr[2];
				eiAddr->oct3 = priv->myRowHead->etherAddr[3];
				eiAddr->oct4 = priv->myRowHead->etherAddr[4];
				eiAddr->oct5 = priv->myRowHead->etherAddr[5];

				/* Send message to this node */
				NG_SEND_MSG_ID (error, node, msg, NG_PEER_NODE(hook)->nd_ID, 0);
			}

			/* Accept the connection */
			return (0);
		}
		else
		{
			/* For the other two hooks, always accept connections */
			return (0);
		}
}


/*
 * Receive a control message
 */
static int
ng_daphne_rcvmsg(node_p node, item_p item, hook_p lasthook)
{
	const priv_p priv = NG_NODE_PRIVATE(node);
	struct ng_mesg *resp = NULL;
	int error = 0;
	struct ng_mesg *msg;

	NGI_GET_MSG(item, msg);
	switch (msg->header.typecookie)
	{
	case NGM_DAPHNE_COOKIE:
		switch (msg->header.cmd)
		{
		case NGM_DAPHNE_SET_ENADDR:
			/* Asked to set a new hardware address to our node */
			if (msg->header.arglen != ETHER_ADDR_LEN) {
				error = EINVAL;
				break;
			}
			memcpy (priv->myRowHead->etherAddr, (u_char *)msg->data, ETHER_ADDR_LEN);
			break;

		case NGM_DAPHNE_GET_ENADDR:
			/* Asked about our current hardware address */
			NG_MKRESPONSE(resp, msg, ETHER_ADDR_LEN, M_NOWAIT);
			if (resp == NULL) {
				error = ENOMEM;
				break;
			}
			bcopy(priv->myRowHead->etherAddr, resp->data, ETHER_ADDR_LEN);
			break;

		default:
			/* If not one of the above, the message was invalid */
			error = EINVAL;
			break;
		}
		break;

	case NGM_ETHER_COOKIE:
		/* Check if this message is the response of the one sent at the
		   'connect' function. This is the only way we accepted a message
		   from a ng_ether node */
		if (!(msg->header.flags & NGF_RESP))
		{
			error = EINVAL;
			break;
		}
		switch (msg->header.cmd)
		{
		case NGM_ETHER_GET_ENADDR:
			/* Construct a rowHead for ourselves. Enter the just accepted hardware
			   address in it. Initialize the other values */
			DEQUEUE_ELEM(priv->spareRowHeads, priv->myRowHead, priv->numSpareRowHeads);

			priv->myRowHead->dataSeqNum = 1;
			priv->myRowHead->routingSeqNum = 1;
			memcpy (priv->myRowHead->etherAddr, msg->data, 6);

			/* Insert this rowHead to the routing table list */
			ADD_TO_CHAIN(priv->routingTable, priv->myRowHead, priv->numRowHeads);

			/* Now that we know our address, we can start the timers and bcast updates */
			callout_reset(&priv->handle, BEACON_PERIOD * hz, ng_daphne_timeout, node);
			break;

		default:
			error = EINVAL;
			break;
		}
		break;

	default:
		error = EINVAL;
		break;
	}

	/* Done */
	NG_RESPOND_MSG(error, node, item, resp);
	NG_FREE_MSG(msg);
	return (error);
}


/*
 * Receive a data packet
 */
static int
ng_daphne_rcvdata(hook_p hook, item_p item)
{
	struct ether_header		*eh;
	struct daphne_header	*dh;
	const node_p 			 node = NG_HOOK_NODE (hook);
	const priv_p 			 priv = NG_NODE_PRIVATE (node);
	int 					 error = 0;
	struct mbuf				*m;

	NGI_GET_M (item, m);

	/* We received this packet from an ether device ? */
	if (hook == priv->downlink)
	{
		/* Sanity check */
		if ((m->m_len < ETHER_HDR_LEN) && ((m = m_pullup(m, ETHER_HDR_LEN)) == NULL))
		{
			NG_FREE_ITEM (item);
			NG_FREE_M (m);
			return (EINVAL);
		}

		eh = mtod (m, struct ether_header *);

		/* According to its ethertype, is it of type daphne ? */
		if (eh->ether_type != htons (DAPHNE_ETHER_TYPE))
		{
			/* It is not daphne, pass it through as is */
			if (priv->pass)
			{
				NG_FWD_NEW_DATA (error, item, priv->pass, m);
			}
			else
			{
				NG_FREE_ITEM (item);
				NG_FREE_M (m);
				error = 0;
			}
			return (error);
		}
		else
		{
			/* If needed, pull up the packet so that the daphne header is in one piece */
			if ((m->m_len < ETHER_HDR_LEN + DAPHNE_HDR_LEN) &&
				((m = m_pullup(m, ETHER_HDR_LEN + DAPHNE_HDR_LEN)) == NULL))
			{
				/* Pull up needed, but failed. Something definitely wrong */
				NG_FREE_ITEM (item);
				NG_FREE_M (m);
				return (EINVAL);
			}

			eh = mtod(m, struct ether_header *);
			dh = (struct daphne_header *)(eh + 1);

			/* Someone sent us a beacon */
			if (dh->type == DAPHNE_BEACON)
			{
				error = ng_daphne_process_beacon (priv, item, m);
				return (error);
			}

			/* Someone sent us a routing update */
			if (dh->type == DAPHNE_ROUTING)
			{
				error = ng_daphne_process_update (priv, item, m);
				return (error);
			}

			/* Someone sent us data encapsulated according to daphne */
			if (dh->type == DAPHNE_DATA)
			{
				error = ng_daphne_process_data_from_downlink (priv, item, m);
				return (error);
			}
		}
	}
	else if (hook == priv->uplink)
		{
			/* Process a packet coming from an upper layer */
			error = ng_daphne_process_data_from_uplink (priv, item, m);
			return (error);
		}
		else if (hook == priv->info)
			{
				/* We do not accept data from the info hook */
				NG_FREE_ITEM (item);
				NG_FREE_M (m);
				return (0);
			}
			else if (hook == priv->pass)
			{
				if (priv->downlink)
				{
					NG_FWD_NEW_DATA (error, item, priv->downlink, m);
				}
				else
				{
					NG_FREE_ITEM (item);
					NG_FREE_M (m);
					error = 0;
				}
				return (error);
			}
			else
			{
				/* Where did it come from? Let's panic */
				panic(__FUNCTION__);
			}

	return (error);
}


/*
 * Shutdown node
 */
static int
ng_daphne_shutdown(node_p node)
{
	const priv_p priv = NG_NODE_PRIVATE(node);
	struct rowHead *p_tempRowHead;

	/*
	 * Shut down everything except the timer. There's no way to
	 * avoid another possible timeout event (it may have already
	 * been dequeued), so we can't free the node yet.
	 */

	 /* Tear down the routing table, prepare to shutdown */
	for (p_tempRowHead = priv->routingTable; p_tempRowHead != NULL; p_tempRowHead = p_tempRowHead->next)
	{
		FREE_CHAIN (p_tempRowHead->neighbours, rtElem);
	}
	FREE_CHAIN(priv->routingTable, rowHead);

	priv->routingTable = NULL;
	priv->myRowHead = NULL;
	priv->numRowHeads = 0;

	/* NG_INVALID flag is now set so node will be freed at next timeout */
	return (0);
}


/*
 * Disconnect a hook
 */
static int
ng_daphne_disconnect(hook_p hook)
{
	const priv_p priv = NG_NODE_PRIVATE(NG_HOOK_NODE(hook));

	/* Accept hook disconnection without any actions */
	if (hook == priv->downlink)
	{
		priv->downlink = NULL;
	}
	else if (hook == priv->uplink)
		{
			priv->uplink = NULL;
		}
		else if (hook == priv->info)
			{
				priv->info = NULL;
			}
			else if (hook == priv->pass)
			{
				priv->pass = NULL;
			}
			else
			{
				panic(__FUNCTION__);
			}

	/* If no more hooks, go away */
	if ((NG_NODE_NUMHOOKS(NG_HOOK_NODE(hook)) == 0)
	&& (NG_NODE_IS_VALID(NG_HOOK_NODE(hook)))) {
		ng_rmnode_self(NG_HOOK_NODE(hook));
	}

	return (0);
}


/*
 * Handle timer timeouts
 */
static void
ng_daphne_timeout (void *arg)
{
	const node_p node = arg;
	const priv_p priv = NG_NODE_PRIVATE(node);
	int s, calcPathsNeeded = 0;
	struct rowHead	*p_tempRowHead;
	struct rtElem	*p_tempRtElem, *p_tempRtElem2;

	/* We do not want to get interrupted now. The routing table should stay
	   constant throughout the update or beacon bcast process. */
	s = splnet();

	/* If node was shut down, this is the final lingering timeout */
	if (NG_NODE_NOT_VALID(node)) {
		/* Free all memory allocated by now */
		FREE_CHAIN (priv->heardNeighbours, rtElem);
		priv->numHeardNeighbours = 0;

		FREE_CHAIN (priv->spareRtElems, rtElem);
		priv->numSpareRtElems = 0;

		FREE_CHAIN (priv->spareRowHeads, rowHead);
		priv->numSpareRowHeads = 0;

		FREE(priv, M_NETGRAPH_DAPHNE);
		NG_NODE_SET_PRIVATE(node, NULL);
		NG_NODE_UNREF(node);
		splx(s);
		return;
	}

	/* Do we need more spare rtElems? */
	if (priv->numSpareRtElems < MIN_SPARE_RTELEMS)
	{
		MALLOC(p_tempRtElem, struct rtElem*, sizeof (struct rtElem), M_NETGRAPH_DAPHNE, M_NOWAIT);
		if (p_tempRtElem)
		{
			ADD_TO_CHAIN(priv->spareRtElems, p_tempRtElem, priv->numSpareRtElems);
		}
	}
	else
		/* Or are there too many ? */
		if (priv->numSpareRtElems > MAX_SPARE_RTELEMS)
		{
			DEQUEUE_ELEM(priv->spareRtElems, p_tempRtElem, priv->numSpareRtElems);
			FREE(p_tempRtElem, M_NETGRAPH_DAPHNE);
		}

	/* Do we need more spare rowHeads? */
	if (priv->numSpareRowHeads < MIN_SPARE_ROWHEADS)
	{
		MALLOC(p_tempRowHead, struct rowHead *, sizeof (struct rowHead), M_NETGRAPH_DAPHNE, M_NOWAIT);
		if (p_tempRowHead)
		{
			ADD_TO_CHAIN(priv->spareRowHeads, p_tempRowHead, priv->numSpareRowHeads);
		}
	}
	else
		/* Or are there too many ? */
		if (priv->numSpareRowHeads > MAX_SPARE_ROWHEADS)
		{
			DEQUEUE_ELEM(priv->spareRowHeads, p_tempRowHead, priv->numSpareRowHeads);
			FREE(p_tempRowHead, M_NETGRAPH_DAPHNE);
		}

	/* Check for heard neighbours that may have timed_out */
	if (priv->heardNeighbours)
	{
		p_tempRtElem = priv->heardNeighbours;
		while (p_tempRtElem->next)
		{
			p_tempRtElem->next->timeOut--;
			if (p_tempRtElem->next->timeOut == 0)
			{
				p_tempRtElem2 = p_tempRtElem->next;
				p_tempRtElem->next = p_tempRtElem2->next;
				priv->numHeardNeighbours--;
				ADD_TO_CHAIN(priv->spareRtElems, p_tempRtElem2, priv->numSpareRtElems);
			}
			else
			{
				p_tempRtElem = p_tempRtElem->next;
			}
		}
		p_tempRtElem = priv->heardNeighbours;
		p_tempRtElem->timeOut--;
		if (p_tempRtElem->timeOut == 0)
		{
			priv->heardNeighbours = p_tempRtElem->next;
			ADD_TO_CHAIN(priv->spareRtElems, p_tempRtElem, priv->numSpareRtElems);
			priv->numHeardNeighbours--;
		}
	}

	/* Check for confirmed neighbours that may have timed out */
	if (priv->myRowHead->neighbours)
	{
		p_tempRtElem = priv->myRowHead->neighbours;
		while (p_tempRtElem->next)
		{
			p_tempRtElem->next->timeOut--;
			if (p_tempRtElem->next->timeOut == 0)
			{
				p_tempRtElem2 = p_tempRtElem->next;
				p_tempRtElem->next = p_tempRtElem2->next;
				priv->myRowHead->numNeighb--;
				ADD_TO_CHAIN(priv->spareRtElems, p_tempRtElem2, priv->numSpareRtElems);
				calcPathsNeeded = 1;
			}
			else
			{
				p_tempRtElem = p_tempRtElem->next;
			}
		}
		p_tempRtElem = priv->myRowHead->neighbours;
		p_tempRtElem->timeOut--;
		if (p_tempRtElem->timeOut == 0)
		{
			priv->myRowHead->neighbours = p_tempRtElem->next;
			ADD_TO_CHAIN(priv->spareRtElems, p_tempRtElem, priv->numSpareRtElems);
			priv->myRowHead->numNeighb--;
			calcPathsNeeded = 1;
		}
	}

	/* Check for sequence numbers that may have timed out. We use different sequence numbers
	   for data packets and for routning updates. If timed out, set corresponding seqnum to 0
	   (wildcard). Do not check for ourselves, we never time out */
	for (p_tempRowHead = priv->routingTable; p_tempRowHead != priv->myRowHead; p_tempRowHead = p_tempRowHead->next)
	{
		/* Check for routing seqnum */
		if (p_tempRowHead->routingSeqNumTimer > 0)
		{
			p_tempRowHead->routingSeqNumTimer--;
			if (p_tempRowHead->routingSeqNumTimer == 0)
			{
				p_tempRowHead->routingSeqNum = 0;
			}
		}

		/* Check for data seqnum */
		if (p_tempRowHead->dataSeqNumTimer > 0)
		{
			p_tempRowHead->dataSeqNumTimer--;
			if (p_tempRowHead->dataSeqNumTimer == 0)
			{
				p_tempRowHead->dataSeqNum = 0;
			}
		}
	}

	/* Issue a new sequence number */
	priv->myRowHead->routingSeqNum++;

	/* 0 is never used as a valid seqnum */
	if (priv->myRowHead->routingSeqNum == 0)
	{
		priv->myRowHead->routingSeqNum ++;
	}

	/* Is it time to send a full update? */
	if (priv->updateCounter % BEACON_ROUTING_RATIO == 0)
	{
		/* Yes it is */
		ng_daphne_transmit_update (priv);
		if (priv->info != NULL)
		{
			/* If info hook connected, send the routing table on this hook */
			ng_daphne_transmit_rt_info (priv);
		}
		priv->updateCounter = 1;
	}
	else
	{
		/* No, send only a beacon */
		ng_daphne_transmit_beacon (priv);
		priv->updateCounter ++;
	}

	if (calcPathsNeeded)
	{
		ng_daphne_calc_shortest_paths (priv);
	}

	/* Renew the timer */
	callout_reset(&priv->handle, BEACON_PERIOD * hz, ng_daphne_timeout, node);

	splx(s);
}


/*
 * Transmit a beacon
 */
static int
ng_daphne_transmit_beacon (priv_p priv)
{
	const u_char 			 ether_bcast_addr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	struct ether_header		*eh;
	struct daphne_header	*dh;
	struct mbuf				*m;
	int 					 error = 0;
	u_int32_t				 size;
	struct rtElem			*p_tempRtElem;
	u_int8_t				*p_work;
	meta_p					 meta = NULL;

	/* Total size of the beacon packet */
	size = ETHER_HDR_LEN + DAPHNE_HDR_LEN + ETHER_ADDR_LEN * priv->numHeardNeighbours + 1;

	/* Get an mbuf */
	MGETHDR (m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
	{
		return (ENOBUFS);
	}

	/* If mbuf too small, get a cluster */
	if (size > MHLEN)
	{
		MCLGET (m, M_DONTWAIT);
		if ((M_EXT & m->m_flags) == 0)
		{
			m_freem(m);
			return (ENOBUFS);
		}
	}

	m->m_pkthdr.len = m->m_len = size;
	m->m_pkthdr.rcvif = NULL;

	/* Fill up ether header */
	eh = (struct ether_header *)m->m_data;
	memcpy (eh->ether_dhost, ether_bcast_addr, ETHER_ADDR_LEN);
	memcpy (eh->ether_shost, priv->myRowHead->etherAddr, ETHER_ADDR_LEN);
	eh->ether_type = htons(DAPHNE_ETHER_TYPE);

	/* Fill up daphne header */
	dh = (struct daphne_header *) (eh + 1);
	memcpy (dh->ether_dhost, ether_bcast_addr, ETHER_ADDR_LEN);
	memcpy (dh->ether_shost, priv->myRowHead->etherAddr, ETHER_ADDR_LEN);
	dh->type = DAPHNE_BEACON;
	dh->seq_num = htonl (priv->myRowHead->routingSeqNum);
	dh->ttl = 1;
	dh->size = htons (size - ETHER_HDR_LEN);
	dh->orig_type = htons (DAPHNE_ETHER_TYPE);

	p_work = (u_int8_t *)(dh + 1);

	/* Insert number of heard neighbours contained in this beacon */
	*(p_work++) = priv->numHeardNeighbours;

	/* Insert all heard neighbours into packet */
	for (p_tempRtElem = priv->heardNeighbours; p_tempRtElem != NULL; p_tempRtElem = p_tempRtElem->next)
	{
		memcpy (p_work, p_tempRtElem->etherAddr, ETHER_ADDR_LEN);
		p_work += ETHER_ADDR_LEN;
	}

	/* We are ready, send the packet */
	NG_SEND_DATA (error, priv->downlink, m, meta);

	return (error);
}


/*
 * Transmit a routing update
 */
static int
ng_daphne_transmit_update (priv_p priv)
{
	const u_char 			 ether_bcast_addr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	struct ether_header		*eh;
	struct daphne_header	*dh;
	struct mbuf				*m;
	int 					 error = 0;
	u_int32_t				 residualSize, rowHeadSize, rowHeadsIncluded, updateSize;
	struct rtElem			*p_tempRtElem;
	struct rowHead			*p_tempRowHead;
	u_int8_t				*p_work;
	meta_p					 meta = NULL;

	p_tempRowHead = priv->routingTable;

	/* Go to the first entry that should be added to this update */
	while ((p_tempRowHead != NULL) && (p_tempRowHead->routingSeqNum == 0))
	{
		p_tempRowHead = p_tempRowHead->next;
	}

newFragment:

	/* Get an mbuf */
	MGETHDR (m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
	{
		return (ENOBUFS);
	}

	/* Always add a cluster. Usually it is going to be needed */
	MCLGET (m, M_DONTWAIT);
	if ((M_EXT & m->m_flags) == 0)
	{
		m_freem(m);
		return (ENOBUFS);
	}

	m->m_pkthdr.rcvif = NULL;

	/* Fill up ether header */
	eh = (struct ether_header *)m->m_data;
	memcpy (eh->ether_dhost, ether_bcast_addr, ETHER_ADDR_LEN);
	memcpy (eh->ether_shost, priv->myRowHead->etherAddr, ETHER_ADDR_LEN);
	eh->ether_type = htons(DAPHNE_ETHER_TYPE);

	/* Fill up daphne header */
	dh = (struct daphne_header *) (eh + 1);
	memcpy (dh->ether_dhost, ether_bcast_addr, ETHER_ADDR_LEN);
	memcpy (dh->ether_shost, priv->myRowHead->etherAddr, ETHER_ADDR_LEN);
	dh->type = DAPHNE_ROUTING;
	dh->seq_num = htonl (priv->myRowHead->routingSeqNum);
	dh->ttl = 1;
	dh->orig_type = htons (DAPHNE_ETHER_TYPE);

	/* Leave one octet free for number of rowHeads Included */
	p_work = (u_int8_t *)(dh + 1) + 1;

	/* Choose update size: max cluster size or fixed upper limit? */
	residualSize = (MCLBYTES < MAX_UPDATE_SIZE) ? MCLBYTES : MAX_UPDATE_SIZE;

	/* We have already used some packet size. Substract it */
	residualSize -= sizeof (struct daphne_header) + sizeof (struct ether_header) + 1;

	rowHeadsIncluded = 0;

newRowHead:

	/* How much room is needed for this rowHead ? */
	rowHeadSize = (p_tempRowHead->numNeighb + 1) * ETHER_ADDR_LEN +
		sizeof(p_tempRowHead->numNeighb) + sizeof(p_tempRowHead->routingSeqNum);

	/* Is there enough room in the current packet ? */
	if (residualSize > rowHeadSize)
	{
		rowHeadsIncluded++;

		/* Substract rowHead size from available size in packet */
		residualSize -= rowHeadSize;

		/* Add rowHead in packet */
		memcpy (p_work, p_tempRowHead->etherAddr, ETHER_ADDR_LEN);
		p_work += ETHER_ADDR_LEN;
		*((u_int32_t *)p_work) = htonl (p_tempRowHead->routingSeqNum);
		p_work += sizeof(p_tempRowHead->routingSeqNum);
		*(p_work++) = p_tempRowHead->numNeighb;

		for (p_tempRtElem = p_tempRowHead->neighbours; p_tempRtElem != NULL;
			p_tempRtElem = p_tempRtElem->next)
		{
			memcpy (p_work, p_tempRtElem->etherAddr, ETHER_ADDR_LEN);
			p_work += ETHER_ADDR_LEN;
		}
	}
	else
	{
		/* There is not enough room in current packet */

		/* Mark in packet how many rowHeads are included */
		p_work = (u_int8_t *)(dh + 1);
		*p_work = rowHeadsIncluded;

		/* Insert size in mbuf and daphne header */
		updateSize = (MCLBYTES < MAX_UPDATE_SIZE ? MCLBYTES : MAX_UPDATE_SIZE) -
			residualSize - ETHER_HDR_LEN;
		m->m_pkthdr.len = m->m_len = updateSize + ETHER_HDR_LEN;
		dh->size = htons (updateSize);

		/* Send this packet */
		NG_SEND_DATA (error, priv->downlink, m, meta);

		/* We need one more packet */
		goto newFragment;
	}

	/* Go to the next packet that should be added to this update */
	do
	{
		p_tempRowHead = p_tempRowHead->next;
	} while ((p_tempRowHead != NULL) && (p_tempRowHead->routingSeqNum == 0));

	/* Is there another rowHead ? */
	if (p_tempRowHead != NULL)
	{
		/* Yes, add it to packet */
		goto newRowHead;
	}
	else
	{
		/* No, we are done */

		/* Mark in packet how many rowHeads are included */
		p_work = (u_int8_t *)(dh + 1);
		*p_work = rowHeadsIncluded;

		/* Insert size in mbuf and daphne header */
		updateSize = (MCLBYTES < MAX_UPDATE_SIZE ? MCLBYTES : MAX_UPDATE_SIZE) -
			residualSize - ETHER_HDR_LEN;
		m->m_pkthdr.len = m->m_len = updateSize + ETHER_HDR_LEN;
		dh->size = htons (updateSize);

		/* Send our final packet */
		NG_SEND_DATA (error, priv->downlink, m, meta);
	}

	return (error);
}


/*
 * Process a received beacon
 */
static int
ng_daphne_process_beacon (priv_p priv, struct ng_item *item, struct mbuf *m)
{
	u_int8_t				*tempBuffer, *p_work;
	u_int8_t				 numberOfEntries, i;
	u_int32_t				 bufferSize;
	struct daphne_header	*dh;

	tempBuffer = NULL;

	/* Allocate memory for a buffer */
	bufferSize = m->m_pkthdr.len;
	MALLOC (tempBuffer, u_int8_t *, bufferSize, M_NETGRAPH_DAPHNE, M_NOWAIT);
	if (tempBuffer == NULL)
	{
		NG_FREE_ITEM(item);
		NG_FREE_M (m);
		return (ENOMEM);
	}

	/* Copy packet contents to buffer */
	m_copydata (m, 0, bufferSize, tempBuffer);

	/* The packet is not needed any more */
	NG_FREE_ITEM(item);
	NG_FREE_M (m);

	dh = (struct daphne_header *)(tempBuffer + ETHER_HDR_LEN);

	/* Add node that sent this beacon in heard neighbours list */
	ng_daphne_add_heard_neighbour (priv, dh->ether_shost);

	p_work = (u_int8_t *)(dh + 1);
	numberOfEntries = *(p_work++);

	for (i = 0; i < numberOfEntries; i++)
	{
		/* If I am included in the heard neighbours list of the node that
		   sent the beacon, then this node is a confirmed neighbour (Two
		   way link exists) */
		if (ETHER_EQUAL(priv->myRowHead->etherAddr, p_work))
		{
			ng_daphne_add_neighbour (priv, dh->ether_shost);
		}
		p_work += ETHER_ADDR_LEN;
	}

	FREE (tempBuffer, M_NETGRAPH_DAPHNE);
	return (0);
}


/*
 * Process a received routing update
 */
static int
ng_daphne_process_update (priv_p priv, struct ng_item *item, struct mbuf *m)
{
	u_int8_t				*tempBuffer, *p_work;
	u_int8_t				 numberOfEntries, tempNumNeighb, i, j;
	u_int32_t				 bufferSize, tempSeqNum;
	struct daphne_header	*dh;
	struct rowHead			*p_tempRowHead;
	struct rtElem			*p_tempRtElem;

	tempBuffer = NULL;

	/* Allocate memory for a buffer */
	bufferSize = m->m_pkthdr.len;
	MALLOC (tempBuffer, u_int8_t *, bufferSize, M_NETGRAPH_DAPHNE, M_NOWAIT);
	if (tempBuffer == NULL)
	{
		NG_FREE_ITEM(item);
		NG_FREE_M (m);
		return (ENOMEM);
	}

	/* Copy packet contents into buffer */
	m_copydata (m, 0, bufferSize, tempBuffer);

	/* The packet is not needed any more */
	NG_FREE_ITEM(item);
	NG_FREE_M(m);

	dh = (struct daphne_header *)(tempBuffer + ETHER_HDR_LEN);

	/* Add node that sent the update to heard neighbours list */
	ng_daphne_add_heard_neighbour (priv, dh->ether_shost);

	p_work = (u_int8_t *)(dh + 1);
	numberOfEntries = *(p_work++);

	/* For all the rowheads included in this update */
	for (i = 0; i < numberOfEntries; i++)
	{
		FIND_ROWHEAD(p_tempRowHead, p_work);

		/* There is no entry for this node ? */
		if (p_tempRowHead == NULL)
		{
			/* Dequeue or malloc a rowHead for this node */
			DEQUEUE_ELEM(priv->spareRowHeads, p_tempRowHead, priv->numSpareRowHeads);
			if (p_tempRowHead == NULL)
			{
				MALLOC(p_tempRowHead, struct rowHead *, sizeof (struct rowHead), M_NETGRAPH_DAPHNE, M_NOWAIT);
				if (p_tempRowHead == NULL)
				{
					FREE(tempBuffer, M_NETGRAPH_DAPHNE);
					return (ENOMEM);
				}
			}

			/* Initialize it and insert the node's address */
			bzero(p_tempRowHead, sizeof (struct rowHead));
			memcpy (p_tempRowHead->etherAddr, p_work, ETHER_ADDR_LEN);
			p_tempRowHead->numHops = DAPHNE_INFINITY;

			/* Add it to the routing table */
			ADD_TO_CHAIN(priv->routingTable, p_tempRowHead, priv->numRowHeads);
		}

		p_work += ETHER_ADDR_LEN;

		tempSeqNum = ntohl (*((u_int32_t *)p_work));
		p_work += 4;
		tempNumNeighb = *(p_work++);


		if ((p_tempRowHead->routingSeqNum == 0) || (SEQ_GT(tempSeqNum, p_tempRowHead->routingSeqNum)))
		{
			/* If the routing seqnum is 0 (wildcard) or the packet contains a fresher seqnum, update
			   its contents */

			p_tempRowHead->routingSeqNum = tempSeqNum;
			p_tempRowHead->routingSeqNumTimer = TICKS_UNTIL_ROUTING_TO;

			/* If there are recorded neighbours in this rowHead, erase them */
			if (p_tempRowHead->neighbours != NULL)
			{
				for (p_tempRtElem = p_tempRowHead->neighbours; p_tempRtElem->next != NULL;
					p_tempRtElem = p_tempRtElem->next)
					;

				p_tempRtElem->next = priv->spareRtElems;
				priv->spareRtElems = p_tempRowHead->neighbours;
				priv->numSpareRtElems += p_tempRowHead->numNeighb;

				p_tempRowHead->neighbours = NULL;
				p_tempRowHead->numNeighb = 0;
			}

			/* Add new neighbours to this rowHead, as per the received update */
			for (j = 0; j < tempNumNeighb; j++)
			{
				DEQUEUE_ELEM(priv->spareRtElems, p_tempRtElem, priv->numSpareRtElems);
				if (p_tempRtElem == NULL)
				{
					MALLOC (p_tempRtElem, struct rtElem *, sizeof (struct rtElem), M_NETGRAPH_DAPHNE, M_NOWAIT);
					if (p_tempRtElem == NULL)
					{
						FREE_CHAIN (p_tempRowHead->neighbours, rtElem);
						p_tempRowHead->neighbours = NULL;
						p_tempRowHead->numNeighb = NULL;
						p_tempRowHead->routingSeqNum = 0;
						p_tempRowHead->routingSeqNumTimer = 0;

						FREE(tempBuffer, M_NETGRAPH_DAPHNE);
						return(ENOMEM);
					}
				}

				memcpy (p_tempRtElem->etherAddr, p_work, ETHER_ADDR_LEN);
				p_work += ETHER_ADDR_LEN;

				ADD_TO_CHAIN(p_tempRowHead->neighbours, p_tempRtElem, p_tempRowHead->numNeighb);
			}
		}
		else
		{
			/* Go to the next rowHead in the packet */
			p_work += tempNumNeighb * ETHER_ADDR_LEN;
		}
	}

	FREE(tempBuffer, M_NETGRAPH_DAPHNE);

	/* Routing table may have changed. Recalculate the shortest paths */
	ng_daphne_calc_shortest_paths (priv);

	return (0);
}


/*
 * Calculate shortest paths
 */
static int
ng_daphne_calc_shortest_paths (priv_p priv)
{
	struct rowHead	*p_tempRowHead, *p_tempRowHead2;
	struct rtElem	*p_tempRtElem;
	u_int8_t		 workingHop, finished;

	/* Set all nodes except me as being unreachable */
	for (p_tempRowHead = priv->routingTable; p_tempRowHead != priv->myRowHead;
		p_tempRowHead = p_tempRowHead->next)
	{
		p_tempRowHead->numHops = DAPHNE_INFINITY;
	}

	/* Set that my neighbours are 1 hop away and they are the next hop to themselves */
	for (p_tempRtElem = priv->myRowHead->neighbours; p_tempRtElem != NULL;
		p_tempRtElem = p_tempRtElem->next)
	{
		FIND_ROWHEAD(p_tempRowHead, p_tempRtElem->etherAddr);
		if (p_tempRowHead == NULL)
		{
			DEQUEUE_ELEM(priv->spareRowHeads, p_tempRowHead, priv->numSpareRowHeads);
			if (p_tempRowHead == NULL)
			{
				MALLOC(p_tempRowHead, struct rowHead *, sizeof (struct rowHead), M_NETGRAPH_DAPHNE, M_NOWAIT);
				if (p_tempRowHead == NULL)
				{
					return (ENOMEM);
				}
			}
			bzero(p_tempRowHead, sizeof (struct rowHead));
			memcpy (p_tempRowHead->etherAddr, p_tempRtElem->etherAddr, ETHER_ADDR_LEN);
			ADD_TO_CHAIN(priv->routingTable, p_tempRowHead, priv->numRowHeads);
		}

		p_tempRowHead->numHops = 1;
		memcpy (p_tempRowHead->nextHop, p_tempRowHead->etherAddr, ETHER_ADDR_LEN);
	}

	workingHop = 1;
	finished = 0;

	while (finished == 0)
	{
		finished = 1;

		/* Check all rowHeads in the routing table */
		for (p_tempRowHead = priv->routingTable; p_tempRowHead != NULL;
			p_tempRowHead = p_tempRowHead->next)
		{
			/* Find the ones that are as many hops away as we are working on now */
			if (p_tempRowHead->numHops == workingHop)
			{
				/* Find their neighbours */
				for (p_tempRtElem = p_tempRowHead->neighbours; p_tempRtElem != NULL;
					p_tempRtElem = p_tempRtElem->next)
				{
					/* Add the corresponding rowHead if it does not exist */
					FIND_ROWHEAD(p_tempRowHead2, p_tempRtElem->etherAddr);
					if (p_tempRowHead2 == NULL)
					{
						DEQUEUE_ELEM(priv->spareRowHeads, p_tempRowHead2, priv->numSpareRowHeads);
						if (p_tempRowHead2 == NULL)
						{
							MALLOC(p_tempRowHead2, struct rowHead *, sizeof (struct rowHead),
								M_NETGRAPH_DAPHNE, M_NOWAIT);

							if (p_tempRowHead2 == NULL)
							{
								return (ENOMEM);
							}
						}
						bzero(p_tempRowHead2, sizeof (struct rowHead));
						memcpy (p_tempRowHead2->etherAddr, p_tempRtElem->etherAddr, ETHER_ADDR_LEN);
						p_tempRowHead2->numHops = DAPHNE_INFINITY;
						ADD_TO_CHAIN(priv->routingTable, p_tempRowHead2, priv->numRowHeads);
					}

					/* Did we find a shorter route ? */
					if (p_tempRowHead2->numHops > workingHop)
					{
						/* Yes, update numHops and nextHop entries */
						p_tempRowHead2->numHops = workingHop + 1;
						memcpy (p_tempRowHead2->nextHop, p_tempRowHead->nextHop, ETHER_ADDR_LEN);

						/* We are not finished yet, make another iteration */
						finished = 0;
					}
				}
			}
		}
		workingHop ++;
	}

	for (p_tempRowHead = priv->routingTable; p_tempRowHead != NULL; p_tempRowHead=p_tempRowHead->next)
	{
		printf ("%02x:%02x : %d\n", p_tempRowHead->etherAddr[4], p_tempRowHead->etherAddr[5], p_tempRowHead->numHops);
	}

	/* If there are still nodes that are unreachable, delete them */

	/* Besides me, are there any more nodes in the routing table ? */
	if (priv->routingTable != priv->myRowHead)
	{
		p_tempRowHead = priv->routingTable;
		while (p_tempRowHead->next)
		{
			printf ("%d\n", p_tempRowHead->next->numHops);
			if (p_tempRowHead->next->numHops == DAPHNE_INFINITY)
			{
				p_tempRowHead2 = p_tempRowHead->next;
				p_tempRowHead->next = p_tempRowHead2->next;
				priv->numRowHeads--;
				ADD_TO_CHAIN(priv->spareRowHeads, p_tempRowHead2, priv->numSpareRowHeads);
			}
			else
			{
				p_tempRowHead = p_tempRowHead->next;
			}
		}
		p_tempRowHead = priv->routingTable;
		if (p_tempRowHead->numHops == DAPHNE_INFINITY)
		{
			printf ("first: %d\n", p_tempRowHead->numHops);
			priv->routingTable = p_tempRowHead->next;
			ADD_TO_CHAIN(priv->spareRowHeads, p_tempRowHead, priv->numSpareRowHeads);
			priv->numRowHeads--;
		}
	}

	return (0);
}

/*
 * Receive a data packet on the downlink
 */
static int
ng_daphne_process_data_from_downlink (priv_p priv, struct ng_item *item, struct mbuf *m)
{
	struct ether_header		*eh;
	struct daphne_header	*dh;
	struct rowHead			*p_tempRowHead;
	const u_char			 ether_bcast_addr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	u_int8_t				 ttl;
	struct mbuf				*m2 = NULL;
	meta_p					 meta, meta2 = NULL;
	int						 error;

	eh = mtod(m, struct ether_header *);
	dh = (struct daphne_header *)(eh + 1);

	/* Decrease the time to live value by one and update the packet daphne header.
	   If negative, discard the packet */
	ttl = dh->ttl - 1;
	if (ttl > 0)
	{
		dh->ttl = ttl;
	}
	else
	{
		NG_FREE_ITEM(item);
		NG_FREE_M(m);
		return (0);
	}

	/* Check for broadcast or unicast packet */
	if (!memcmp(dh->ether_dhost, ether_bcast_addr, ETHER_ADDR_LEN))
	{
		/* We received a broadcast message */

		/* Is the source of this message in our routing table ? */
		FIND_ROWHEAD(p_tempRowHead, dh->ether_shost);
		if (p_tempRowHead == NULL)
		{
			/* If not make and initialize an entry for it */
			DEQUEUE_ELEM(priv->spareRowHeads, p_tempRowHead, priv->numSpareRowHeads);
			if (p_tempRowHead == NULL)
			{
				MALLOC(p_tempRowHead, struct rowHead *, sizeof (struct rowHead), M_NETGRAPH_DAPHNE, M_NOWAIT);
				if (p_tempRowHead == NULL)
				{
					NG_FREE_ITEM(item);
					NG_FREE_M(m);
					return (ENOMEM);
				}
			}
			bzero(p_tempRowHead, sizeof (struct rowHead));
			memcpy (p_tempRowHead->etherAddr, dh->ether_shost, ETHER_ADDR_LEN);
			p_tempRowHead->numHops = DAPHNE_INFINITY;
			ADD_TO_CHAIN(priv->routingTable, p_tempRowHead, priv->numRowHeads);

			/* Routing Table Actions complete, relay packet */
			goto relay_bcast_message;
		}
		else
		{
			/* Check if packet contains fresher seqnum, or the rowhead contains wildcard */
			if ((SEQ_GT(ntohl(dh->seq_num), p_tempRowHead->dataSeqNum)) || (p_tempRowHead->dataSeqNum == 0))
			{
				goto relay_bcast_message;
			}
			else
			{
				NG_FREE_ITEM(item);
				NG_FREE_M(m);
				return (0);
			}
		}
	}
	else
	{
		/* We received a unicast message */
		if (!memcmp(dh->ether_dhost, priv->myRowHead->etherAddr, ETHER_ADDR_LEN))
		{
			/* It is for me */
			//printf ("Packet for me!!!\n");
			m_adj(m, DAPHNE_HDR_LEN);
			NG_FWD_NEW_DATA(error, item, priv->uplink, m);
			return (error);
		}
		else
		{
			/* Who is it for ? */
			FIND_ROWHEAD(p_tempRowHead, dh->ether_dhost);

			/* If unknown, drop packet */
			if (p_tempRowHead == NULL)
			{
				NG_FREE_ITEM(item);
				NG_FREE_M(m);
				return (EINVAL);
			}
			else
			{
				/* If known but unreachable, drop packet */
				if (p_tempRowHead->numHops == DAPHNE_INFINITY)
				{
					NG_FREE_ITEM(item);
					NG_FREE_M(m);
					return (EINVAL);
				}

				memcpy (eh->ether_shost, priv->myRowHead->etherAddr, ETHER_ADDR_LEN);
				memcpy (eh->ether_dhost, p_tempRowHead->nextHop, ETHER_ADDR_LEN);

				NG_FWD_NEW_DATA(error, item, priv->downlink, m);
				return (error);
			}
		}
	}

relay_bcast_message:

	p_tempRowHead->dataSeqNum = ntohl (dh->seq_num);
	p_tempRowHead->dataSeqNumTimer = TICKS_UNTIL_DATA_TO;

	NGI_GET_META (item, meta);

	/* Clone packet */
	m2 = m_dup (m, M_NOWAIT);

	if (m2 == NULL)
	{
		NG_FREE_ITEM (item);
		NG_FREE_M (m);
		NG_FREE_META (meta);
		return (ENOBUFS);
	}

	if ((meta != NULL) && ((meta2 = ng_copy_meta (meta)) == NULL))
	{
		NG_FREE_ITEM (item);
		NG_FREE_M (m);
		NG_FREE_META (meta);
		m_freem(m2);
		return (ENOMEM);
	}

	/* Send one copy without the daphne header upwards */
	m_adj(m2, DAPHNE_HDR_LEN);
	NG_SEND_DATA (error, priv->uplink, m2, meta2);

	/* Send the other copy downwards, with my ether address as source */
	memcpy (eh->ether_shost, priv->myRowHead->etherAddr, ETHER_ADDR_LEN);
	NG_SEND_DATA (error, priv->downlink, m, meta);

	return (error);
}


/*
 * Received a packet from the upper layers
 */
static int
ng_daphne_process_data_from_uplink (priv_p priv, struct ng_item *item, struct mbuf *m)
{
	struct ether_header		*eh;
	struct daphne_header	*dh;
	struct rowHead			*p_tempRowHead;
	const u_char			 ether_bcast_addr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	u_int16_t				 packetSize;
	int						 error;

	/* Make a sanity check */
	if ((m->m_len < ETHER_HDR_LEN) && ((m = m_pullup(m, ETHER_HDR_LEN)) == NULL))
	{
		NG_FREE_ITEM (item);
		NG_FREE_M (m);
		return (EINVAL);
	}

	/* Make room for daphne header */
	M_PREPEND(m, DAPHNE_HDR_LEN, M_NOWAIT);

	if (!m)
	{
		NG_FREE_ITEM(item);
		return(ENOBUFS);
	}

	if ((m->m_len < (ETHER_HDR_LEN + DAPHNE_HDR_LEN)) &&
		((m = m_pullup(m, (ETHER_HDR_LEN + DAPHNE_HDR_LEN))) == NULL))
	{
		NG_FREE_ITEM (item);
		NG_FREE_M (m);
		return (EINVAL);
	}

	eh = mtod (m, struct ether_header *);
	dh = (struct daphne_header *)(eh + 1);

	/* Update packet size in mbuf */
	packetSize = m->m_pkthdr.len - ETHER_HDR_LEN;

	/* Fill up daphne header */
	dh->type = DAPHNE_DATA;
	dh->ttl = DAPHNE_TTL;
	dh->size = packetSize;

	memcpy (eh->ether_shost, dh->ether_shost, ETHER_ADDR_LEN);
	eh->ether_type = htons(DAPHNE_ETHER_TYPE);

	if (!memcmp (dh->ether_dhost, ether_bcast_addr, ETHER_ADDR_LEN))
	{
		/* Broadcast message */
		memcpy (eh->ether_dhost, ether_bcast_addr, ETHER_ADDR_LEN);

		if ((++(priv->myRowHead->dataSeqNum)) == 0)
		{
			priv->myRowHead->dataSeqNum = 1;
		}

		dh->seq_num = priv->myRowHead->dataSeqNum;

		NG_FWD_NEW_DATA(error, item, priv->downlink, m);
		return (error);
	}
	else
	{
		/* Unicast message */

		/* Who is it destined for ? */
		FIND_ROWHEAD(p_tempRowHead, dh->ether_dhost);

		if (p_tempRowHead == NULL)
		{
			/* If unknown drop the packet */
			NG_FREE_ITEM (item);
			NG_FREE_M(m);
			return (0);
		}
		else
		{
			if (p_tempRowHead->numHops == DAPHNE_INFINITY)
			{
				/* If known, but unreachable, drop the packet */
				NG_FREE_ITEM (item);
				NG_FREE_M (m);
				return (0);
			}
			else
			{
				/* Copy next hop to the ethernet destination and send packet */
				memcpy(eh->ether_dhost, p_tempRowHead->nextHop, ETHER_ADDR_LEN);
				NG_FWD_NEW_DATA(error, item, priv->downlink, m);
				return (error);
			}
		}
	}
}


/*
 * Add a heard neighbour
 */
static int
ng_daphne_add_heard_neighbour (priv_p priv, u_int8_t *addr)
{
	struct rtElem	*p_tempRtElem;

	for (p_tempRtElem = priv->heardNeighbours; p_tempRtElem != NULL; p_tempRtElem = p_tempRtElem->next)
	{
		if (ETHER_EQUAL(addr, p_tempRtElem->etherAddr))
		{
			/* Entry exists, just reset its timer */
			p_tempRtElem->timeOut = BEACON_MISSES_TILL_TO;
			break;
		}
	}

	if (p_tempRtElem != NULL)
	{
		return (0);
	}
	else
	{
		/* Entry does not exist, initialize a rowHead and add it to the
		   heard neighbours list */
		DEQUEUE_ELEM (priv->spareRtElems, p_tempRtElem, priv->numSpareRtElems);
		if (p_tempRtElem == NULL)
		{
			MALLOC(p_tempRtElem, struct rtElem *, sizeof (struct rtElem), M_NETGRAPH_DAPHNE, M_NOWAIT);
		}

		if (p_tempRtElem == NULL)
		{
			return (ENOMEM);
		}

		memcpy (p_tempRtElem->etherAddr, addr, ETHER_ADDR_LEN);
		p_tempRtElem->timeOut = BEACON_MISSES_TILL_TO;
		ADD_TO_CHAIN (priv->heardNeighbours, p_tempRtElem, priv->numHeardNeighbours);
		return (0);
	}
}


static int
ng_daphne_add_neighbour (priv_p priv, u_int8_t *addr)
{
	struct rtElem	*p_tempRtElem;

	for (p_tempRtElem = priv->myRowHead->neighbours; p_tempRtElem != NULL; p_tempRtElem = p_tempRtElem->next)
	{
		if (ETHER_EQUAL(addr, p_tempRtElem->etherAddr))
		{
			/* Entry exists, just reset its timer */
			p_tempRtElem->timeOut = BEACON_MISSES_TILL_TO;
			break;
		}
	}

	if (p_tempRtElem != NULL)
	{
		return (0);
	}
	else
	{
		/* Entry does not exist, initialize a rowHead and add it to the
		   neighbours list */
		DEQUEUE_ELEM (priv->spareRtElems, p_tempRtElem, priv->numSpareRtElems);
		if (p_tempRtElem == NULL)
		{
			MALLOC(p_tempRtElem, struct rtElem *, sizeof (struct rtElem), M_NETGRAPH_DAPHNE, M_NOWAIT);
		}

		if (p_tempRtElem == NULL)
		{
			return (ENOMEM);
		}

		memcpy (p_tempRtElem->etherAddr, addr, ETHER_ADDR_LEN);
		p_tempRtElem->timeOut = BEACON_MISSES_TILL_TO;
		ADD_TO_CHAIN (priv->myRowHead->neighbours, p_tempRtElem, priv->myRowHead->numNeighb);
		return (0);
	}
}

/*
 * Send routing table to info hook
 */
static int
ng_daphne_transmit_rt_info (priv_p priv)
{
	struct mbuf		*m, *m2, *m3;
	meta_p			 meta = NULL;
	u_int32_t		totalSize, currentSize, rowHeadSize;
	u_char			*workingPointer;
	struct rowHead	*pRowHead;
	struct rtElem	*pRtElem;
	int				 error;

	/* Get an mbuf */
	MGETHDR (m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
	{
		return (ENOBUFS);
	}

	/* and add a cluster to it */
	MCLGET (m, M_DONTWAIT);
	if ((M_EXT & m->m_flags) == 0)
	{
		m_freem (m);
		return (ENOBUFS);
	}

	m->m_pkthdr.rcvif = NULL;
	totalSize = 3;
	currentSize = 3;
	rowHeadSize = 0;
	pRowHead = priv->routingTable;

	workingPointer = (u_char *)m->m_data;
	m2 = m;

	/* This is of type INFO_ROUTING_TABLE */
	*(workingPointer++) = INFO_ROUTING_TABLE;

	/* How many rowHeads will be included in the message ? */
	*((u_int16_t *)workingPointer) = priv->numRowHeads;
	workingPointer += 2;

	while (pRowHead != NULL)
	{
		rowHeadSize = (pRowHead->numNeighb + 1) * 6 + 1;

		if (rowHeadSize + currentSize > MCLBYTES)
		{
			/* We need another mbuf */
			MGET (m3, M_DONTWAIT, MT_DATA);

			/* If MGET failed, tear down the mbuf chain */
			if (m3 == NULL)
			{
				do
				{
					m = m_free (m);
				} while (m != NULL);
				return (ENOBUFS);
			}

			/* Add a cluster to the mbuf */
			MCLGET (m3, M_DONTWAIT);

			/* If could not add cluster, tear down the mbuf chain */
			if ((M_EXT & m3->m_flags) == 0)
			{
				do
				{
					m = m_free (m);
				} while (m != NULL);
				return (ENOBUFS);
			}

			m2->m_len = currentSize;

			m2->m_next = m3;
			m2 = m3;
			currentSize = 0;
			workingPointer = (u_char *)m2->m_data;
		}

		memcpy (workingPointer, pRowHead->etherAddr, 6);
		workingPointer += 6;
		*(workingPointer++) = pRowHead->numNeighb;
		for (pRtElem = pRowHead->neighbours; pRtElem != NULL; pRtElem = pRtElem->next)
		{
			memcpy (workingPointer, pRtElem->etherAddr, 6);
			workingPointer += 6;
		}
		totalSize += rowHeadSize;
		currentSize += rowHeadSize;

		pRowHead = pRowHead->next;
	}

	m2->m_len = currentSize;

	m->m_pkthdr.len = totalSize;

	NG_SEND_DATA (error, priv->info, m, meta);
	return (error);
}

